#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Une proposition de correction du TP d'intégration numérique

"""

from random import random
from math import pi
import matplotlib.pyplot as plt
import numpy as np

##########################################################################
#                              PARTIE 1                                  #
##########################################################################


def carre(x):
    return(x**2)

def integ_rectG(f,a,b,n):
    # f est une fonction, a<b des réels, n>0 un entier.
    # renvoit l'approximation de l'intégrale de f entre a et b obtenue 
    # avec n rectangles à gauche
    pas=(b-a)/n
    integrale=0
    for i in range(n):
        integrale=integrale+pas*f(a+i*pas)
    return(integrale)
    

def integ_pointM(f,a,b,n):
    # f est une fonction, a<b des réels, n>0 un entier.
    # renvoit l'approximation de l'intégrale de f entre a et b obtenue 
    # avec la méthode du point milieu
    pas=(b-a)/n
    integrale=0
    for i in range(n):
        integrale=integrale+pas*f(a+i*pas+pas/2)
    return(integrale)
    
    
def integ_trapeze(f,a,b,n):
    # f est une fonction, a<b des réels, n>0 un entier.
    # renvoit l'approximation de l'intégrale de f entre a et b obtenue 
    # avec la méthode des trapezes (naïf : pas optimisé sur les calculs d'images)
    pas=(b-a)/n
    integrale=0
    for i in range(n):
        integrale=integrale+pas*(f(a+i*pas)+f(a+(i+1)*pas))*.5
    return(integrale)


# ci-dessous, les méthodes élémentaires et la fonction qui crée la méthode composée.

def integ_elem_rectG(f,a,b):
    return(f(a)*(b-a))

def integ_elem_mil(f,a,b):
    return(f((a+b)/2)*(b-a))

def integ_elem_trapeze(f,a,b):
    return((f(a)+f(b))/2*(b-a))
    
def integ_comp(methelementaire,f,a,b,n):
    pas=(b-a)/n
    approx=0
    for i in range(n):
        approx=approx+methelementaire(f,a+i*pas,a+(i+1)*pas)
    return(approx)
        
    
def integ_monte_carlo(f,a,b,n):
    # f une fonction intégrable sur [a;b] , n>0 un entier naturel
    # renvoie une approximation de l'intégrale de f sur [a;b] obtenue par
    # la méthode de Monte-Carlo
    valMoy=0
    for i in range(n):
        alea=a+(b-a)*random()
        valMoy=valMoy+f(alea)
    valMoy=valMoy/n
    return(valMoy*(b-a))
    
def Runge(x):
    #fonction de Runge : x -> 1/(25+x^2)
    return(1/(25+x**2))


valeurExacte=pi/20

def compareInteg():
    # provoque l'affichage d'un graphique des courbes des erreurs des différentes
    # méthodes d'intégration en fonction du nombre de points utilisés.
    N=[10,20,30,50,100,150,300,500,1000,1500,2000,5000,10000]
    erreur_RectG=[]
    erreur_pointM=[]
    erreur_Trapeze=[]
    erreur_MonteCarlo=[]
    
    for n in N :
        erreur_RectG.append(abs(valeurExacte-integ_rectG(Runge,0,5,n)))
        erreur_pointM.append(abs(valeurExacte-integ_pointM(Runge,0,5,n)))
        erreur_Trapeze.append(abs(valeurExacte-integ_trapeze(Runge,0,5,n)))
        erreur_MonteCarlo.append(abs(valeurExacte-integ_monte_carlo(Runge,0,5,n)))
           
    plt.plot(N,erreur_RectG,'-b',label='rectangles à gauche')
    plt.plot(N,erreur_pointM,'-c',label='point milieu')
    plt.plot(N,erreur_Trapeze,'-g',label='trapèzes')
    plt.plot(N,erreur_MonteCarlo,'-r',label='Monte Carlo')
    
    plt.legend( loc='upper right', numpoints = 1 )
    plt.xlabel('nombre de points')
    plt.ylabel('\'erreur')
    plt.show()
    
    
def compareIntegLogLog():
    # provoque l'affichage d'un graphique des courbes des erreurs des différentes
    # méthodes d'intégration en fonction du nombre de points utilisés.
    N=[10,20,30,50,100,150,300,500,1000,1500,2000,5000,10000]
    erreur_RectG=[]
    erreur_pointM=[]
    erreur_Trapeze=[]
    erreur_MonteCarlo=[]
    
    for n in N :
        erreur_RectG.append(abs(valeurExacte-integ_rectG(Runge,0,5,n)))
        erreur_pointM.append(abs(valeurExacte-integ_pointM(Runge,0,5,n)))
        erreur_Trapeze.append(abs(valeurExacte-integ_trapeze(Runge,0,5,n)))
        erreur_MonteCarlo.append(abs(valeurExacte-integ_monte_carlo(Runge,0,5,n)))
        
    plt.plot(np.log(N),np.log(erreur_RectG),'-b',label='rectangles à gauche')
    plt.plot(np.log(N),np.log(erreur_pointM),'-c',label='point milieu')
    plt.plot(np.log(N),np.log(erreur_Trapeze),'-g',label='trapèzes')
    plt.plot(np.log(N),np.log(erreur_MonteCarlo),'-r',label='Monte Carlo')
    
    plt.legend( loc='lower left', numpoints = 1 )
    plt.xlabel('logarithme du nombre de points')
    plt.ylabel('logarithme de l\'erreur')
    plt.show()
    
    
# pour les rectangles, l'erreur est en O(1/n) ce qui donne ln(erreur)=a-ln(n) : on a donc les points sur une droite de pente -1
# pour le point milieu et les trapezes, l'erreur est en O(1/n^2) ce qui donne ln(erreur)=a-2ln(n) : on a donc les points sur une droite de pente -2
# Monte Carlo a une erreur qui est de l'ordre de 1/sqrt(n) (d'où une vague pente -1/2) mais l'aléatoire fait que les points ne sont pas alignés 
    

##########################################################################
#                              PARTIE 2                                  #
##########################################################################


def integ_elem_simpson(f,a,b):
    return((f(a)+f(b)+4*f((a+b)/2))/6*(b-a))
    
def compareIntegLogLog2():
    # provoque l'affichage d'un graphique des courbes des erreurs des différentes
    # méthodes d'intégration en fonction du nombre de points utilisés.
    N=[10,20,30,50,100,150,300,500,1000]
    erreur_RectG=[]
    erreur_pointM=[]
    erreur_Trapeze=[]
    erreur_Simpson=[]
    
    for n in N :
        erreur_RectG.append(abs(valeurExacte-integ_rectG(Runge,0,5,n)))
        erreur_pointM.append(abs(valeurExacte-integ_pointM(Runge,0,5,n)))
        erreur_Trapeze.append(abs(valeurExacte-integ_trapeze(Runge,0,5,n)))
        erreur_Simpson.append(abs(valeurExacte-integ_comp(integ_elem_simpson,Runge,0,5,n)))
    
    plt.plot(np.log(N),np.log(erreur_RectG),'-b',label='rectangles à gauche')
    plt.plot(np.log(N),np.log(erreur_pointM),'-c',label='point milieu')
    plt.plot(np.log(N),np.log(erreur_Trapeze),'-g',label='trapèzes')
    plt.plot(np.log(N),np.log(erreur_Simpson),'-r',label='Simpson')
    
    plt.legend( loc='lower right', numpoints = 1 )
    plt.xlabel('logarithme du nombre de points')
    plt.ylabel('logarithme de l\'erreur')
    plt.show()
    
    
# Lorsque l'erreur devient de l'ordre de 10^(-16), on atteint les limites du type float. Autrement dit  :
# le type float ne permet plus de faire la différence entre les nombres. Ceci explique l'allure de la courbe.
    
    

##########################################################################
#                              PARTIE 3                                  #
##########################################################################

def estDansS(x,y):
    # (x,y) sont les coordonnées d'un point du plan
    # la fonction renvoie True si le point considéré est dans la surface 
    # S:x^2+3y^2<=1, False sinon.
    return(x**2+3*y**2<=1)
    
def approxS(n):
    # n est un int
    # renvoie l'approximation obtenue avec la méthode de Monte Carlo pour S
    nombre_dans_S=0
    for i in range(n):
        x=2*random()-1
        y=2*random()-1
        if estDansS(x,y):
            nombre_dans_S=nombre_dans_S+1
            
    proportion=nombre_dans_S/n
    
    return(proportion*4)    
    
# on observe une assez forte instabilité du résultat : avec 10^4 points, on a seulement 2 décimales
# qui sont "stables". On peut un peu améliorer les choses en prenant un rectangle plus petit qui contient
# S (sur l'axe des ordonnées, on peut varier uniquement entre +/- 1/sqrt(3))
   
   
def approxS_bis(n):
    # n est un int
    # renvoie l'approximation obtenue avec la méthode de Monte Carlo pour S
    nombre_dans_S=0
    inv_racine3=3**(-.5)
    for i in range(n):
        x=2*random()-1
        y=2*inv_racine3*random()-inv_racine3
        if estDansS(x,y):
            nombre_dans_S=nombre_dans_S+1
            
    proportion=nombre_dans_S/n
    
    return(proportion*4*inv_racine3)    
    
